/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * 8080-emulator.c
 * Copyright (C) 2010 Ole Reinhardt <ole.reinhardt@embedded-it.de>
 *                    Thermotemp GmbH
 * 
 */

#include <sys/types.h>
#include <sys/timer.h>
#include <sys/stat.h>
#include <sys/heap.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>

#include <cfg/mmci.h>
#include <dev/board.h>
#include <dev/reset.h>
#include <dev/mmcard.h>
#include <fs/phatfs.h>

void run_8080(int, char**, char*);

/* If your architecture does not yet support SPI bus MMC devices, define the
   block device driver to use below.
 */

#if defined(ETHERNUT3)

/* Ethernut 3 block device interface. */
#define MY_BLKDEV      devNplMmc0
#define MY_BLKDEV_NAME "MMC0"

#elif defined(AT91SAM9260_EK)

/* SAM9260-EK block device interface. */
#define MY_BLKDEV      devAt91Mci0
#define MY_BLKDEV_NAME "MCI0"

#endif


static FILE *init_uart(NUTDEVICE *dev, char *name)
{
    uint32_t baud = 115200;
    uint32_t stop   = 1;
    uint32_t data   = 8;
    uint32_t parity = 0;
    
    FILE *uart;

    if (NutRegisterDevice(dev, 0, 0) != 0) {
        return NULL;
    }

    uart = fopen(name, "r+");

    _ioctl(_fileno(uart), UART_SETSPEED,    &baud);
    _ioctl(_fileno(uart), UART_SETSTOPBITS, &stop);
    _ioctl(_fileno(uart), UART_SETPARITY,   &parity);
    _ioctl(_fileno(uart), UART_SETDATABITS, &data);

    return uart;
}

static int init_mmc(void)
{  
    int mmc_volume;
    /* Register the PHAT file system. */
    if (NutRegisterDevice(&devPhat0, 0, 0)) {
        fprintf(stderr, "Registering PHAT: failed\r\n");        
        return -1;
    } else {
        fprintf(stderr, "Registering PHAT: OK\r\n");  
    }

#ifdef MY_BLKDEV
    /* Most architectures still use direct MMC block devices, not the 
       SPI Bus MMC devices. Define the block device for your board above...
     */
    
    if (NutRegisterDevice(&MY_BLKDEV, 0, 0)) {
        fprintf(stderr, "Registering block device '" MY_BLKDEV_NAME "' failed.\r\n");
        return -1;
    } else {
        fprintf(stderr, "Registering block device '" MY_BLKDEV_NAME "' OK.\r\n");
    }

#else         
    /* Register the MMC block device as SPI Bus client device. This is the more 
       modern way used e.g. on eNet-sam7X board and is available for Ethernut 3 
       as well. 
     */
    if (NutRegisterSpiDevice(&DEV_MMCARD0_SPIBUS, &DEV_SPIBUS0, SPI_MMC_CS)) {
        fprintf(stderr, "Registering %s: failed\r\n", DEV_MMCARD0_NAME);        
        return -1;        
    } else {
        fprintf(stderr, "Registering %s: OK\r\n", DEV_MMCARD0_NAME);        
    }
#endif    
    
    /* Mount partition. */
    fprintf(stderr, "Mounting partition: \r\n");
    
    if ((mmc_volume = _open(DEV_MMCARD0_NAME":1/PHAT0", _O_RDWR | _O_BINARY)) == -1) {
        fprintf(stderr, "Mounting partition: failed\r\n");
        return -1;
    } else {
        printf("Mounting partition: success\r\n");
    }
    
    return 0;
}


/* Reads line from standard input. */
static int edit_line(char *prompt, char *line, int siz)
{
    int ch;
    int pos = strlen(line);
 
    printf("%s: %s", prompt, line);
    fflush(stdout);
    for (;;) {
        ch = getchar();
        if (ch == 8) {
            if (pos) {
                pos--;
                printf("\b \b");
            }
        }
        else if (ch < ' ') {
            break;
        }
        else if (pos + 1 < siz) {
            putchar(ch);
            line[pos++] = ch;
        }
        else {
            putchar('\a');
        }
    }
    line[pos] = 0;
    putchar('\n');
 
    return 0;
}

static void show_rootdir(void)
{
    DIR *dir;
    struct dirent *dir_entry;
    int  count = 0;

    printf("\r\n\r\nFiles on MMC-Card:\r\n");
    printf("Nr:   Filename\r\n\r\n");
    
    dir = opendir("PHAT0:/");

    while ((dir_entry = readdir(dir)) != NULL) {
        printf("%3d : %s\r\n", count, dir_entry->d_name);
        count ++;
    }

    closedir(dir);
    printf("\n");
}

int main(void)
{
    FILE  *uart0;
    char   ch;
    char   import_filename[16] = "";
    char   import_filepath[32] = "";
    int    argument_count = 0;
    char  *bios_filename;
    char  *cp;
    char  *arg[2] = {NULL, NULL};
    struct stat stat_buf;

    /* Initialise the UART and stdio */
    uart0 = init_uart(&DEV_UART0, DEV_UART0_NAME);
    
    freopen(DEV_UART0_NAME, "wb", stdout);
    freopen(DEV_UART0_NAME, "rb", stdin);
    freopen(DEV_UART0_NAME, "wb", stderr);

    /* Clear screen */
    printf("\e[2J");    
    printf("\r\nIntel 8080 CP/M Emulation. Running at %ldMHz, %d bytes free\r\n", 
           NutGetCpuClock() / 1000000, NutHeapAvailable());
    
    fprintf(stderr,"Registering MMC...\r\n");
    if (init_mmc() != 0) {
        fprintf(stderr,"ERROR: MMC failed...\r\n");
        goto halt;
    }    

    printf("\r\nStartup options. Please select 1, 2 or 3\r\n\r\n");
    printf("1 -- Start emulator (run BASIC interpreter bios)\r\n");
    printf("2 -- Start emulator (run CP/M bios)\r\n");
    printf("3 -- Select file to import to CP/M disk image, then start emulator (CP/M bios)\r\n");

    do {
        ch = getchar();
    } while ((ch != '1') && (ch != '2') && (ch != '3'));

    switch (ch) {
        case '1':
            bios_filename = "PHAT0:/c_basic.bin";
            break;
            
        case '2':
            bios_filename = "PHAT0:/c_bios.bin";
            break;
            
        case '3':
            bios_filename = "PHAT0:/c_bios.bin";
            /* Show MMC cards root directory */
            show_rootdir();

            /* Ask for a file to import into a CP/M disk image */
            printf("Do you want to import any files? (just press enter if not)\r\n");
            edit_line("Import? > ", import_filename, sizeof(import_filename));

            /* Strip off trailing linefeed */
            cp = strchr(import_filename, '\n');
            if (cp) {
                *cp = 0;
            }

            if (strlen(import_filename) > 0) {
                /* Prepend 'PHAT0:/' to the filename */
                strcpy(import_filepath, "PHAT0:/");
                strcat(import_filepath, import_filename);

                /* Check if the given file exists */
                if (stat(import_filepath, &stat_buf) == 0) {
                    printf("\r\nCall \"IMPORT %s\" in the CP/M system to import \"%s\"\r\n", 
                           import_filename, import_filepath);
                    arg[1] = import_filepath;
                    argument_count = 1;
                } else {
                    printf("\r\nInvalid filename: %s\r\n", import_filepath);
                }

                NutSleep(500);
            } else {
                printf("\r\n");
            }
            break;
    }

    if (stat(bios_filename, &stat_buf) != 0) {
        printf("\r\nBIOS image \"%s\" not found\r\n", bios_filename);
        goto halt;
    }
    
    
    printf("Starting emulator...\r\n");
    NutSleep(1000);
    /* Clear screen */
    printf("\e[2J");
    run_8080(argument_count + 1, arg, bios_filename);

halt:    
    printf("\r\nEmulator halted, restart? [Y/N]\r\n");
    ch = getchar();
    if ((ch == 'y') || (ch == 'Y')) {
        /* Reset the board */
        NutReset();
    }
    
    for (;;) {
        NutSleep(500);
    }
    return 0;
}
